/* Process.c
 * Windows2000/XP Process Command for Tcl. 32bit Only.
 * This Program doesn't work on Windows95/98/ME
 *
 * Copyright (C) 2005 Reddog <reddog@s35.xrea.com>
 * 
 * Redistribution  and  use  in  source  and  binary  forms,  with  or  without
 * modification, are permitted provided  that the following conditions are met:
 * (1) Redistributions of  source code must retain the  above copyright notice,
 * this list  of conditions and the following  disclaimer.  (2) Redistributions
 * in  binary form  must reproduce  the above  copyright notice,  this  list of
 * conditions and  the following disclaimer  in the documentation  and/or other
 * materials provided  with the distribution.  (3)  The name of  any author may
 * not  be used  to  endorse or  promote  products derived  from this  software
 * without their specific prior written permission.
 *
 * THIS SOFTWARE  IS PROVIDED  "AS IS" AND  ANY EXPRESS OR  IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
 * AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
 * SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
 * INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
 * CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
 * ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include <windows.h>
#include <tlhelp32.h>
#include "psapi.h"
#include <tcl.h>

/*#define __IA64__*/
/*#define USE_NAMESPACE*/
/*#define DEBUG*/

#define VERSION "1.0"

#ifdef USE_NAMESPACE
#define NAMESPACE "process::"
#else
#define NAMESPACE
#endif

#define OBJEQSTR(obj, str) (strcmp(Tcl_GetString((Tcl_Obj*)(obj)), (str)))

typedef struct _TProcessItem {
	PROCESSENTRY32 pe32;
	char *name; //pointer to pe32.szExeFile
	char path[MAX_PATH];
	int pid;
	HANDLE handle;
	struct {
		FILETIME create;
		FILETIME exit;
		FILETIME kernel;
		FILETIME user;
	} time;
	PROCESS_MEMORY_COUNTERS memory;
} ProcessItem;

typedef struct _TGlobalInfo {
	HANDLE snapshot;
	FILETIME snaptime;
#ifdef __IA64__
	MEMORYSTATUSEX memory;
#else
	MEMORYSTATUS memory;
#endif
} GlobalInfo;

int Tcl_WrongArgs(Tcl_Interp *interp, Tcl_Obj *commandNameObj, char *sample)
{/*{{{*/
	Tcl_Obj *wr = Tcl_NewStringObj("wrong # args: should be \"", -1);
	Tcl_AppendToObj(wr, Tcl_GetString(commandNameObj), -1);
	Tcl_AppendToObj(wr, " ",-1);
	Tcl_AppendToObj(wr, sample, -1);
	Tcl_AppendToObj(wr, "\"",-1);
	Tcl_SetObjResult(interp, wr);
	return TCL_ERROR;
}/*}}}*/

// Ansi to Tcl_String
Tcl_Obj *getTclStringObj(LPCSTR in)
{/*{{{*/
	Tcl_Obj *tclStr;
	LPWSTR uniStr;
	int uniLen;
	char  *utfStr;
	int utfLen;

	// Ansi To Unicode
	uniLen = MultiByteToWideChar(CP_ACP, 0, in, -1, NULL, 0);
	if (uniLen == 0) {
		return Tcl_NewStringObj("", -1);
	}
	uniStr = GlobalAlloc(GPTR, uniLen * sizeof(wchar_t));
	MultiByteToWideChar(CP_ACP, 0, in, -1, uniStr, uniLen);

	// To Utf-8
	utfLen = WideCharToMultiByte(CP_UTF8, 0, uniStr, uniLen, NULL, 0, NULL, NULL);
	utfStr = GlobalAlloc(GPTR, utfLen);
	WideCharToMultiByte(CP_UTF8, 0, uniStr, uniLen, utfStr, utfLen, NULL, NULL);

	// To Tcl_String
	tclStr = Tcl_NewStringObj(utfStr, -1);

	GlobalFree(uniStr);
	GlobalFree(utfStr);
	return tclStr;
}/*}}}*/

// type is 'U' or 'D' (User/Domain)
void getAccount(HANDLE hProcess, char type, char *out, int len)
{/*{{{*/
	HANDLE hToken;
	TOKEN_USER pUserInfo[16];
	DWORD returnLen = 0;
	BYTE name[1024];
	BYTE domain[1024];
	DWORD nameSize = sizeof(name);
	DWORD domainSize = sizeof(domain);
	SID_NAME_USE use;

	if (hProcess == NULL) {
		out[0] = '\0'; return;
	}

	if (OpenProcessToken(hProcess, TOKEN_READ, &hToken) == FALSE) {
#ifdef DEBUG
		printf("Err: Cann't open ProcessToken\n");
#endif
		out[0] = '\0'; return;
	}

	if (GetTokenInformation(hToken, TokenUser, pUserInfo, sizeof(pUserInfo), &returnLen) == FALSE) {
#ifdef DEBUG
		printf("Err: Can't get TokenInfo\n");
#endif
		out[0] = '\0'; return;
		return;
	}
	if (LookupAccountSid(NULL, pUserInfo->User.Sid, name, &nameSize, 
			domain, &domainSize, &use)== FALSE) {
#ifdef DEBUG
		printf("Err: Can't get Account From Sid\n");
#endif
		out[0] = '\0'; return;
		return;
	}
	CloseHandle(hToken);
	switch(type) {
		case 'U': strncpy(out, name, len); break;
		case 'D': strncpy(out, domain, len); break;
	}
	return;
}/*}}}*/

void getProcessVersion (int pid, int *major, int *minor)
{/*{{{*/
	DWORD version;
	version = GetProcessVersion(pid);
	if (version == 0) {
		if (major) *major = 0;
		if (minor) *minor = 0;
	}
	if (major) *major = HIWORD(version);
	if (minor) *minor = LOWORD(version);
	return;
}/*}}}*/

__int64 getDateFromFILETIME (FILETIME *time)
{/*{{{*/
	Tcl_WideInt usecSincePosixEpoch;
	LARGE_INTEGER tmp;
	static LARGE_INTEGER posixEpoch;
	posixEpoch.LowPart = 0xD53E8000;
	posixEpoch.HighPart = 0x019DB1DE;

	tmp.LowPart = time->dwLowDateTime;
	tmp.HighPart = time->dwHighDateTime;
	if (tmp.QuadPart == 0) {
		return 0;
	}
	usecSincePosixEpoch = (tmp.QuadPart - posixEpoch.QuadPart)/10;
	return usecSincePosixEpoch / 1000000;
}/*}}}*/

// type is 'R', 'S', 'M' or 'N' (100nanosec/sec/millisec/nanosec)
__int64 getSecondFromFILETIME (FILETIME *time, char type)
{/*{{{*/
	LARGE_INTEGER tmp;
	tmp.LowPart = time->dwLowDateTime;
	tmp.HighPart = time->dwHighDateTime;
	switch (type) {
		case 'R': return tmp.QuadPart; // 100nano sec
		case 'N': return tmp.QuadPart*100; //nano sec
		case 'M': return tmp.QuadPart/10000; //milli sec
		case 'S': return tmp.QuadPart/10000000; //sec
		default: return tmp.QuadPart;
	}
}/*}}}*/

//type is 'T', 'K' or 'U' (total/kernel/user)
double getCpuUsage(GlobalInfo *ginfo, ProcessItem *process, char type)
{/*{{{*/
	_int64 create, kernel, user, snap;
	create = getSecondFromFILETIME(&process->time.create, 'R');
	kernel = getSecondFromFILETIME(&process->time.kernel, 'R');
	user = getSecondFromFILETIME(&process->time.user, 'R');
	snap = getSecondFromFILETIME(&ginfo->snaptime, 'R');

	if (getSecondFromFILETIME(&process->time.create, 'R') == 0)
		return -1;
	switch (type) {
		case 'K':
			return 100.0 * kernel / (double)(snap - create);
		case 'U':
			return 100.0 * user / (double)(snap - create);
		case 'T':
		default:
			return 100.0 * (kernel + user) / (double)(snap - create);
	}
}/*}}}*/

//type is 'T', 'K' or 'U' (total/kernel/user)
void perfCpuUsage(ProcessItem *processList, double *out, int len, int msec, char type)
{/*{{{*/
	int i;
	FILETIME ctime, etime;
	typedef struct _TPrefTime{
		FILETIME kernel;
		FILETIME user;
	} PrefTime;
	PrefTime *st, *ed;
	
	st = GlobalAlloc(GPTR, sizeof(PrefTime)*len);
	ed = GlobalAlloc(GPTR, sizeof(PrefTime)*len);

	for (i=0; i<len; i++) {
		GetProcessTimes(processList[i].handle,
				&ctime, &etime,
				&st[i].kernel, &st[i].user);
	}

	Sleep(msec);

	for (i=0; i<len; i++) {
		GetProcessTimes(processList[i].handle,
				&ctime, &etime,
				&ed[i].kernel, &ed[i].user);
	}
	
	for (i=0; i<len; i++) {
		_int64 stk = getSecondFromFILETIME(&st[i].kernel, 'R');
		_int64 edk = getSecondFromFILETIME(&ed[i].kernel, 'R');
		_int64 stu = getSecondFromFILETIME(&st[i].user, 'R');
		_int64 edu = getSecondFromFILETIME(&ed[i].user, 'R');
		if (processList[i].handle == NULL) {
			out[i] = -1; continue;
		}
		switch (type) {
			case 'K':
				out[i] = (double)(edk - stk) / (msec * 100.0);
			case 'U':
				out[i] = (double)(edu - stu) / (msec * 100.0);
			case 'T':
			default:
				out[i] = (double)(edk + edu - stk - stu) / (msec * 100.0);
		}
	}
	GlobalFree(st);
	GlobalFree(ed);
	return;
}/*}}}*/

void initData (GlobalInfo **global, ProcessItem **processList, int *size)
{/*{{{*/
	int i, len;
	PROCESSENTRY32 pe32;
	FILETIME now;
	HANDLE hSnapshot;
	BOOL result;
	ProcessItem *plist;

	*global = GlobalAlloc(GPTR, sizeof(GlobalInfo));

	hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, (DWORD)NULL);
	GetSystemTimeAsFileTime(&(*global)->snaptime);

#ifdef __IA64__
	GlobalMemoryStatusEx(&(*global)->memory);
#else
	GlobalMemoryStatus(&(*global)->memory);
#endif

	if ((int)hSnapshot == -1) return;

	pe32.dwSize = sizeof (PROCESSENTRY32);

	//Get Process Number 
	result = Process32First(hSnapshot, &pe32);
	len = 1;
	while (result) {
		result = Process32Next(hSnapshot, &pe32);
		len++;
	}
	*size = len;
	
	//Create Process Items Array
	(*global)->snapshot = hSnapshot;
	plist = GlobalAlloc(GPTR, sizeof(ProcessItem)*len);
	i=0;
	result = Process32First(hSnapshot, &pe32);
	while (result) {
		plist[i].pe32 = pe32;
		plist[i].pid = pe32.th32ProcessID;
		plist[i].name = plist[i].pe32.szExeFile;
		plist[i].handle = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, plist[i].pid);

		GetProcessTimes(plist[i].handle, 
				&plist[i].time.create, &plist[i].time.exit,
				&plist[i].time.kernel, &plist[i].time.user);
		
		GetProcessMemoryInfo(plist[i].handle,
				&plist[i].memory, sizeof(plist[i].memory));

		//Find Module Path Name
		{
			MODULEENTRY32 me32 = {sizeof(MODULEENTRY32)};
			HANDLE hModSnap;
			if (pe32.th32ProcessID != 0) {
				hModSnap = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, pe32.th32ProcessID);
				if (Module32First(hModSnap, &me32)) {
					strncpy(plist[i].path, me32.szExePath, sizeof(plist[i].path));
				}
				CloseHandle(hModSnap);
			}
		}

		result = Process32Next(hSnapshot, &pe32);
		i++;
	}
	*processList = plist;
}/*}}}*/

void freeData (GlobalInfo *ginfo, ProcessItem *processList, int len)
{/*{{{*/
	int i;
	for (i=0; i<len; i++) {
		CloseHandle(processList[i].handle);//handle of process
	}
	GlobalFree(processList);
	CloseHandle(ginfo->snapshot);
	GlobalFree(ginfo);
	return;
}/*}}}*/


// Tcl ps Command 
int PsCmd (ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *objv[])
{/*{{{*/
	int c, i;
	int tclResult;
	Tcl_Obj **tclData;
	int processNum;
	ProcessItem *process = NULL;
	GlobalInfo *ginfo = NULL;
	int plainList = FALSE;
	initData(&ginfo, &process, &processNum);

	//Create Tcl Data Elements And Add PID
	tclData = GlobalAlloc(GPTR, sizeof(Tcl_Obj*) * processNum);
	for (i=0; i<processNum; i++) {
		tclData[i] = Tcl_NewListObj(0, NULL);
		Tcl_ListObjAppendElement(interp, tclData[i],
				Tcl_NewIntObj(process[i].pid));
	}

	// Detail Options
	for (c=1; c<objc; c++) {
		//exe name
		if (OBJEQSTR(objv[c], "-exe") == 0) {
			for (i=0; i<processNum; i++) {
				Tcl_ListObjAppendElement(interp, tclData[i],
						getTclStringObj(process[i].name));
			}
			continue;
		}
		//full path name
		if (OBJEQSTR(objv[c], "-path") == 0) {
			char *path;
			int len;
			for (i=0; i<processNum; i++) {
				len = GetLongPathName(process[i].path, NULL, 0);
				path = GlobalAlloc(GPTR, len);
				GetLongPathName(process[i].path, path, len);
				Tcl_ListObjAppendElement(interp, tclData[i],
						getTclStringObj(path));
				GlobalFree(path);
			}
			continue;
		}
		//parent pid
		if (OBJEQSTR(objv[c], "-ppid") == 0) {
			for (i=0; i<processNum; i++) {
				Tcl_ListObjAppendElement(interp, tclData[i], 
						Tcl_NewIntObj(process[i].pe32.th32ParentProcessID));
			}
			continue;
		}
		//priority
		if (OBJEQSTR(objv[c], "-priority") == 0) {
			for (i=0; i<processNum; i++) {
				Tcl_ListObjAppendElement(interp, tclData[i], 
						Tcl_NewIntObj(process[i].pe32.pcPriClassBase));
			}
			continue;
		}
		//handle count
		//if (OBJEQSTR(objv[c], "-handle") == 0) {
		//	for (i=0; i<processNum; i++) {}
		//	continue;
		//}
		//thread count
		if (OBJEQSTR(objv[c], "-thread") == 0) {
			for (i=0; i<processNum; i++) {
				Tcl_ListObjAppendElement(interp, tclData[i], 
						Tcl_NewIntObj(process[i].pe32.cntThreads));
			}
			continue;
		}
		//create date time
		if (OBJEQSTR(objv[c], "-ctime") == 0) {
			for (i=0; i<processNum; i++) {
				Tcl_ListObjAppendElement(interp, tclData[i],
						Tcl_NewWideIntObj(getDateFromFILETIME(&process[i].time.create)));
			}
			continue;
		}
		//kernel time (msec)
		if (OBJEQSTR(objv[c], "-ktime") == 0) {
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewWideIntObj(-1));
				} else {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewWideIntObj(
								getSecondFromFILETIME(&process[i].time.kernel, 'M')));
				}
			}
			continue;
		}
		//user time (msec)
		if (OBJEQSTR(objv[c], "-utime") == 0) {
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewWideIntObj(-1));
				} else {
				Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewWideIntObj(
							getSecondFromFILETIME(&process[i].time.user, 'M')));
				}
			}
			continue;
		}
		//CPU Usage % (total used/total time from created)
		//if next arg is number(msec), re-perform and calc % (used/msec);
		if (OBJEQSTR(objv[c], "-%cpu") == 0) {
			if (c+1<objc) {
				char *nextarg = Tcl_GetString(objv[c+1]);
				if (nextarg[0] != '-') {
					//Re-measure performance per msec
					int msec = atoi(nextarg);
					double *out = GlobalAlloc(GPTR, sizeof(double) * processNum);

					perfCpuUsage(process, out, processNum, msec, 'T');

					for (i=0; i<processNum; i++) {
						if (out[i]<0) {
							Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
						} else {
							Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewDoubleObj(out[i]));
						}
					}
					GlobalFree(out);
					c++;
					continue;
				}
			}
			for (i=0; i<processNum; i++) {
				double pcpu = getCpuUsage(ginfo, &process[i], 'T');
				if (pcpu<0) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
				} else {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewDoubleObj(pcpu));
				}
			}
			continue;
		}
		//Real Memory %
		if (OBJEQSTR(objv[c], "-%size") == 0) {
			unsigned int div = ginfo->memory.dwTotalPhys;
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
				} else {
					__int64 ws = process[i].memory.WorkingSetSize;
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewDoubleObj(100.0 * ws / div));
				}
			}
			continue;
		}
		//Virtual Memory %
		if (OBJEQSTR(objv[c], "-%vsize") == 0) {
			unsigned int div = ginfo->memory.dwTotalPageFile;
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj( -1));
				} else {
					__int64 pf = process[i].memory.PagefileUsage;
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewDoubleObj(100.0 * pf / div));
				}
			}
			continue;
		}
		//WorkingSet Size (byte)
		if (OBJEQSTR(objv[c], "-size") == 0) {
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
				} else {
					Tcl_ListObjAppendElement(interp, tclData[i],
							Tcl_NewIntObj(process[i].memory.WorkingSetSize));
				}
			}
			continue;
		}
		//Peak WorkingSet Size (byte)
		if (OBJEQSTR(objv[c], "-peaksize") == 0) {
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
				} else {
					Tcl_ListObjAppendElement(interp, tclData[i],
							Tcl_NewIntObj(process[i].memory.PeakWorkingSetSize));
				}
			}
			continue;
		}
		//PageFile Usage (byte)
		if (OBJEQSTR(objv[c], "-vsize") == 0) {
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
				} else {
					Tcl_ListObjAppendElement(interp, tclData[i],
							Tcl_NewIntObj(process[i].memory.PagefileUsage));
				}
			}
			continue;
		}
		//Peak PageFile Usage (byte)
		if (OBJEQSTR(objv[c], "-peakvsize") == 0) {
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
				} else {
					Tcl_ListObjAppendElement(interp, tclData[i],
							Tcl_NewIntObj(process[i].memory.PeakPagefileUsage));
				}
			}
			continue;
		}
		//pagefault
		if (OBJEQSTR(objv[c], "-pagefault") == 0) {
			for (i=0; i<processNum; i++) {
				if (process[i].handle == NULL) {
					Tcl_ListObjAppendElement(interp, tclData[i], Tcl_NewIntObj(-1));
				} else {
					Tcl_ListObjAppendElement(interp, tclData[i],
							Tcl_NewIntObj(process[i].memory.PageFaultCount));
				}
			}
			continue;
		}
		//user name
		if (OBJEQSTR(objv[c], "-user") == 0) {
			char buf[50];
			int len = sizeof(buf);
			for (i=0; i<processNum; i++) {
				getAccount(process[i].handle, 'U', buf, len);
				Tcl_ListObjAppendElement(interp, tclData[i],
						getTclStringObj(buf));
			}
			continue;
		}
		//domain name
		if (OBJEQSTR(objv[c], "-domain") == 0) {
			char buf[50];
			int len = sizeof(buf);
			for (i=0; i<processNum; i++) {
				getAccount(process[i].handle, 'D', buf, len);
				Tcl_ListObjAppendElement(interp, tclData[i],
						getTclStringObj(buf));
			}
			continue;
		}
		//process version
		if (OBJEQSTR(objv[c], "-version") == 0) {
			int major, minor;
			char buf[16];
			for (i=0; i<processNum; i++) {
				getProcessVersion(process[i].pid, &major, &minor);
				_snprintf(buf, sizeof(buf), "%d.%d", major, minor);
				Tcl_ListObjAppendElement(interp, tclData[i],
						Tcl_NewStringObj(buf,-1));
			}
			continue;
		}
		
		// output plain list
		if (OBJEQSTR(objv[c], "-plainlist") == 0) {
			plainList = TRUE;
			continue;
		}

		// Wrong Args
		Tcl_WrongArgs(interp, (Tcl_Obj*)objv[0], 
			"?-exe? ?-path? ?-ppid? ?-priority? ?-thread? "
			"?-ctime? ?-ktime? ?-utime? "
			"?-%cpu? ?-%cpu msec? ?-%size? ?-%vsize? "
			"?-size? ?-peaksize? ?-vsize? ?-peakvsize? ?-pagefault? "
			"?-user? ?-domain? ?-version? "
			"?-plainlist? "
		);
		tclResult = TCL_ERROR;
		goto done;
	}
	
	// Output Result
	if (plainList) {
		Tcl_Obj *result = Tcl_NewListObj(0, NULL);
		for (i=0; i<processNum-1; i++) {
			Tcl_ListObjAppendList(interp, result, tclData[i]);
		}
		Tcl_SetObjResult(interp, result);
	} else {
		Tcl_SetObjResult(interp, Tcl_NewListObj(processNum-1, tclData));
	}
	tclResult = TCL_OK;

done:
	freeData(ginfo, process, processNum);
	GlobalFree(tclData);
	return tclResult;
}/*}}}*/


// Tcl kill Command 
int KillCmd (ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *objv[])
{/*{{{*/
	int i, len;
	Tcl_Obj  *tclResult;
	if (objc<=1) return TCL_OK;

	tclResult = Tcl_NewListObj(0, NULL);

	for (i=1; i<objc; i++){
		int pid;
		HANDLE hProcess;
		DWORD exitcode;
		int listlen, k;

		Tcl_ListObjLength(interp, objv[i], &listlen);

		for (k=0; k<listlen; k++) {
			Tcl_Obj *elObj;
			Tcl_ListObjIndex(interp, objv[i], k, &elObj);
			if (Tcl_GetIntFromObj(interp, elObj, &pid) != TCL_OK) {
				Tcl_ListObjAppendElement(interp, tclResult, elObj);
			} else {
				// Try to Kill
				hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, pid);
				if (hProcess == NULL) {
					Tcl_ListObjAppendElement(interp, tclResult, elObj);
					continue;
				}
				GetExitCodeProcess(hProcess, &exitcode);
				if (TerminateProcess(hProcess, exitcode) == FALSE) {
					Tcl_ListObjAppendElement(interp, tclResult, elObj);
				}
				CloseHandle(hProcess);
			}
		}
	}
	Tcl_SetObjResult(interp, tclResult);
	Tcl_ListObjLength(interp, tclResult, &len);
	if (len==0) {
		return TCL_OK;
	} else {
		return TCL_ERROR;
	}
}/*}}}*/


// Entry point
DLLEXPORT int Process_Init (Tcl_Interp *interp)
{/*{{{*/
#ifdef USE_TCL_STUBS
	if (Tcl_InitStubs(interp, "8", 0) == NULL) {
		return TCL_ERROR;
	}
#endif
	Tcl_CreateObjCommand(interp, NAMESPACE "ps", PsCmd, NULL, NULL);
	Tcl_CreateObjCommand(interp, NAMESPACE "kill", KillCmd, NULL, NULL);

	if (Tcl_PkgProvide(interp, "Process", VERSION) != TCL_OK) {
		return TCL_ERROR;
	}
	return TCL_OK;
}/*}}}*/

/* vim: set fdm=marker ts=4 sw=4:*/
